from pymongo import MongoClient
from ukupacha.Utils import Utils
from joblib import Parallel, delayed
import pandas as pd
import numpy as np
import time


class UkuPachaCheckPoint:
    def __init__(self, user="system", password="colavudea", dburi="localhost:1521", mongodb_uri="mongodb://localhost:27017/"):
        self.client = MongoClient(mongodb_uri)
        # Oracle utls functions.
        self.utils = Utils(user=user, password=password, dburi=dburi)

    def create(self, keys: list, oracle_db: str, main_table: str, mongo_db: str, mongo_collection: str):
        # Exmplae query="select COD_RH,COD_PRODUCTO from UDEA_CV.EN_PRODUCTO"
        query = "select "
        for key in keys:
            query += key+","
        query = query[0:-1]
        query += f" from {oracle_db}.{main_table}"
        data = self.utils.request(query)
        data["status"] = np.zeros(data.shape[0], dtype=int)
        self.client[mongo_db][f"{mongo_collection}_checkpoint"].drop()
        self.client[mongo_db][f"{mongo_collection}_checkpoint"].insert_many(
            data.to_dict('records'))

    def exists(self, mongo_db: str, mongo_collection: str):
        ckp_col = f"{mongo_collection}_checkpoint"
        return ckp_col in self.client[mongo_db].list_collection_names()

    def drop(self, mongo_db: str, mongo_collection: str):
        self.client[mongo_db][f"{mongo_collection}_checkpoint"].drop()

    def update(self, mongo_db: str, mongo_collection: str, keys: dict):
        #print(mongo_db, mongo_collection,keys)
        self.client[mongo_db][f"{mongo_collection}_checkpoint"].update_one(
            keys, {"$set": {"status": 1}})

    def get_regs(self, mongo_db: str, mongo_collection: str):
        """
        Function to get registers from the ckp collection that are not already downloaded
        """
        ckp_col = self.client[mongo_db][f"{mongo_collection}_checkpoint"]
        ckpdata = list(ckp_col.find({"status": 0}, {"_id": 0, "status": 0}))
        return ckpdata

    def get_data_chunk(self, oracle_db: str, main_table: str, mongo_db: str, mongo_collection: str, chunk: int, chunk_size: int, ckpdata):
        query = f"select * from {oracle_db}.{main_table} WHERE "
        last_chunk = chunk+chunk_size
        if last_chunk > len(ckpdata):
            last_chunk = chunk + last_chunk - len(ckpdata)
        for row in ckpdata[chunk:last_chunk]:
            req_query = " ("
            for key in row.keys():
                req_query += f" {key}='{row[key]}' AND"
            req_query = req_query[0:-3]
            req_query += ") OR"
            query += req_query
        query = query[0:-3]
        data = self.utils.request(query)
        return data

    def get_data_p(self, oracle_db: str, main_table: str, mongo_db: str, mongo_collection: str, chunk_size=300, jobs=2):
        ckpdata = self.get_regs(mongo_db, mongo_collection)
        chunks = Parallel(n_jobs=jobs, backend='threading', verbose=10)(delayed(self.get_data_chunk)(
            oracle_db, main_table, mongo_db, mongo_collection, chunk, chunk_size, ckpdata) for chunk in range(0, len(ckpdata), chunk_size))
        if len(chunks) > 0:
            data = pd.concat(chunks)
        else:
            data = pd.DataFrame()
        return data

    def get_data(self, oracle_db: str, main_table: str, mongo_db: str, mongo_collection: str, chunk_size=300):
        ckpdata = self.get_regs(mongo_db, mongo_collection)
        chunks = []
        for chunk in range(0, len(ckpdata), chunk_size):
            query = f"select * from {oracle_db}.{main_table} WHERE "
            last_chunk = chunk+chunk_size
            if last_chunk > len(ckpdata):
                last_chunk = chunk + last_chunk - len(ckpdata)
            for row in ckpdata[chunk:last_chunk]:
                req_query = " ("
                for key in row.keys():
                    req_query += f" {key}='{row[key]}' AND"
                req_query = req_query[0:-3]
                req_query += ") OR"
                query += req_query
            query = query[0:-3]
            data = self.utils.request(query)
            chunks.append(data)
        if len(chunks) > 0:
            return pd.concat(chunks)
        else:
            return pd.DataFrame()



import sys
import cx_Oracle
import datetime
import pandas as pd
import json
from bson import ObjectId
from bson import BSONSTR
from bson.codec_options import TypeCodec
from bson.codec_options import TypeRegistry
from bson.codec_options import CodecOptions


class OracleLOBCodec(TypeCodec):
    """
    Class to give support to mongodb to write objects cx_Oracle.LOB to bson strings. 
    """
    python_type = cx_Oracle.LOB    # the Python type acted upon by this type codec
    bson_type = BSONSTR   # the BSON type acted upon by this type codec

    def transform_python(self, value):
        """Function that transforms a custom type value into a type
        that BSON can encode."""
        return ''.join(value.read())

    def transform_bson(self, value):
        """Function that transforms a vanilla BSON type value into our
        custom type."""
        return value


# Creating codec and registry for mongodb
oraclelob_codec = OracleLOBCodec()
oracle_codec_options = CodecOptions(
    type_registry=TypeRegistry([oraclelob_codec]))


class JsonEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for oracle graph,
    all the customized stuff for encoding required for our endpoints
    can be handle in this class
    """

    def default(self, o):
        """
        Method to encodec custom types.
        """
        if isinstance(o, pd.Timestamp):
            return str(o)
        if isinstance(o, type(pd.NaT)):
            return None
        if isinstance(o, cx_Oracle.LOB):
            # https://cx-oracle.readthedocs.io/en/7.1/lob.html#LOB.read
            # WARNING: es posible que no lea todo el contenido,
            # de momento solo esta en pocos campos, no con contenidos muy largos.
            # ex: tabla RE_PROYECTO_INSTITUCION campo NRO_VALOR
            return ''.join(o.read())
        if isinstance(o, datetime.datetime):
            try:
                return datetime.datetime.strftime(o, format='%Y%m%d')
            except ValueError:
                return None
        if isinstance(o, pd.Series):
            return o.to_dict()
        if isinstance(o, ObjectId):
            return str(o)
        return json.JSONEncoder.default(self, o)


class Utils:
    """
    Utility class to handle some Oracle calls.
    """

    def __init__(self, user="system", password="colavudea", dburi="localhost:1521"):
        """
        Constructor to create an Utils Object
        Parameters:
        ----------
        user:str
            Oracle user, by default system, other users can be provided but be sure they have the right permissions in the DB.
        password:str
            Oracle pass for given user
        dburi:str
            Oracle db uri for connector, default "localhost:1521"

        """

        self.connection = cx_Oracle.connect(user=user,
                                            password=password,
                                            dsn=dburi,
                                            threaded=True)

    def request(self, query):
        """
        Perform a request to the Oracle with a query.

        Parameters:
        ----------
        query:str
            SQL query

        Returns:
        ----------
            dataframe with the results
        """
        # https://www.oracle.com/technical-resources/articles/embedded/vasiliev-python-concurrency.html
        # alternavite to evalute if the code above doesnt work
        # https://stackoverflow.com/questions/60887128/how-to-convert-sql-oracle-database-into-a-pandas-dataframe

        try:
            df = pd.read_sql(query, con=self.connection)
        except cx_Oracle.Error as error:
            print(error)
            # if someting is failing with the connector is better to quit.
            sys.exit(1)
        return df

    def get_keys(self, table, ktype="P"):
        """
        Returns the keys from the table on Oracle DB,

        Parameters:
        ----------
        table:str
            table on database ex: EN_PRODUCTO
        ktype:str
            key type, opetions P (Primary), F (Foreing)

        Returns:
        ---------
            pandas dataframe with key information
        """
        query = f"SELECT cols.table_name, cols.column_name, cols.position, cons.status, cons.owner \
                FROM all_constraints cons, all_cons_columns cols WHERE cols.table_name = '{table}' \
                AND cons.constraint_type = '{ktype}' AND cons.constraint_name = cols.constraint_name \
                AND cons.owner = cols.owner  ORDER BY cols.table_name, cols.position"
        return self.request(query)

    def get_tables(self, db):
        """
        Returns the names of the tables available for a given db name.

        Parameters:
        ----------
        db:str
            database name ex: UDEA_CV

        Returns:
        ----------
            list of tables names 
        """
        query = f"SELECT * FROM all_tables WHERE OWNER='{db}'"
        df = self.request(query)
        return list(df["TABLE_NAME"].values)

    def get_db_data(self, db):
        """
        Returns a dictionary where the keys are the tables and the values the dataframes with the data.

        Parameters:
        ----------
        db:str
            Database name

        Returns:
        ----------
            Dictionary with all the data for the given database.
        """
        data = {}
        tables_names = self.get_tables(db)
        for table in tables_names:
            query = f"SELECT * FROM {db}.{table}"
            data[table] = self.request(query)
        return data

    def request_register(self, db, keys, table):
        """
        Returns a register for a given database(or table space), keys(Primary or foreing) and table from Oracle DB,

        Parameters:
        ----------
        db:str
            database for the table (or table space)
        keys:dict
            dictionary with primary or foreing keys and the values for the specific register.
        table:str
            table on database ex: EN_PRODUCTO

        Returns:
        ---------
            pandas dataframe with the information
        """
        query = f"SELECT * FROM {db}.{table} WHERE "
        for key in keys:
            query += f" {key}='{keys[key]}' AND"
        query = query[0:-3]
        req = self.request(query)
        return req


def is_dict(data):
    """
    function to check if data is a dict
    Parameters.
    ----------
    data:any
        it's suppose to be a dict

    Returns:
    ---------
        True if data is a dict otherwise False.
    """
    return isinstance(data, dict)


def is_list(data):
    """
    function to check if data is a list
    Parameters.
    ----------
    data:any
        it's suppose to be a list

    Returns:
    ---------
        True if data is a list otherwise False.
    """
    return isinstance(data, list)


def is_serie(data):
    """
    function to check if data is a pandas serie
    Parameters.
    ----------
    data:any
        it's suppose to be a pandas serie

    Returns:
    ---------
        True if data is a pandas serie otherwise False.
    """
    if is_dict(data) or is_list(data):
        return False
    return True


def section_exist(section, keys):
    """
    Section and subsections means alias in the graph fields,
    ex: the section in the json for EN_RED is network

    some tables doesn´t have sections in the json, for example RE_EVENTO_PROYECTO
    because it is a relationship table.

    Parameters.
    ----------
    section:str
        name of the table
    keys:list
        keys of the graph fields

    Returns:
    ---------
        True if section is found otherwise False.

    """
    for i in list(keys):
        if section == i:
            return True
    return False


def table_exists(fields, table):
    """
    Function to check if the table is in the graph field.

    Parameters.
    ----------
    fields:dict
        graph field with the map of aliases to create the json
    table:str
        name of the table

    Returns:
    ---------
        True if table is found otherwise False.
    """
    for i in list(fields.keys()):
        if table == i:
            return True
    return False


def parse_table(fields, table_name, data_row, filters_function=None):
    """
    Apply the filter and returns a dict,
    table fields to apply aliases to the table columns is not supported anymore,
    the names of the colunms are mapped to the json directly
    """
    data = {}
    # WARNING HERE; AT THE MOMENT I AM NOT PARSING FIELDS WITH ALIAS
    if table_exists(fields, table_name):
        if filters_function:
            data_row = filters_function(table_name, data_row)
        if is_dict(data_row):
            data = data_row
        else:
            data = data_row.to_dict()
    return data


def replace_graph_db_field(graph, value_old, value_new):
    """
    Allows to replace the filed "DB": "__VALUE__" for "DB": "__NEW_VALUE__"
    example for scienti:
        "DB": "__CVLAC__" for "DB": "UDEA_CV"
        "DB": "__GRUPLAC__" for "DB": "UDEA_GR"
        "DB": "__INSTITULAC__" for "DB": "UDEA_IN"

    Parameters:
    ----------
    graph:dict
        dictionary with the model
    value_old:str
        current value for "DB" field
    value_new:str
        new value for "DB" field

    Returns:
    ----------
        dict with the new graph with the field "DB" changed. 
    """
    graph_str = json.dumps(graph)
    graph_str = graph_str.replace(
        f'"DB": "{value_old}"', f'"DB": "{value_new}"')
    return json.loads(graph_str)

# def parse_table(fields,table_name,data_row,remove_nulls=True):
#    data={}
#    if table_exists(fields,table_name):
#        for field,alias in fields[table_name]["fields"].items():
#            if remove_nulls:
#                if data_row[field]:
#                    data[alias] = data_row[field]
#            else:
#                data[alias] = data_row[field]
#    return data



from pymongo import MongoClient
import pandas as pd
from ukupacha.Utils import Utils
from ukupacha.Utils import is_dict, is_list, is_serie, section_exist, table_exists, parse_table, JsonEncoder, oracle_codec_options
from ukupacha.CheckPoint import UkuPachaCheckPoint
from joblib import Parallel, delayed
from tqdm import tqdm
import psutil
import json
import sys


class UkuPachaGraph:
    """
    Class to perform the relations in the database across mutiples tables, databases or table spaces.
    """

    def __init__(self, user="system", password="colavudea", dburi="localhost:1521"):
        """
        Constructor to create an UkuPachaGraph Object
        Parameters:
        ----------
        user:str
            Oracle user, by default system, other users can be provided but be sure they have the right permissions in the DB.
        password:str
            Oracle pass for given user
        dburi:str
            Oracle db uri for connector, default "localhost:1521"

        """
        self.utils = Utils(user=user, password=password, dburi=dburi)

    def request_graph(self, data_row, tables, main_table=None, debug=False):
        """
        Recursive algorithm to walk the graph, performing the requests.
        """
        if debug:
            print("="*30)
        if not data_row.empty:
            ndata = []
            if main_table is None:
                ndata.append(data_row)
            else:
                ndata.append({"table": main_table, "data": [data_row]})

            for table_dict in tables:
                table = list(table_dict.keys())[0]
                if debug:
                    print(f"table = {table}")
                    print(list(table_dict.keys()))
                if table_dict[table] is not None:
                    for table_relations in table_dict[table]:
                        db = table_relations["DB"]
                        keys = {}
                        for key in table_relations["KEYS"]:
                            alias = key
                            if '/' in key:
                                tmp_keys = key.split('/')
                                key = tmp_keys[0]
                                alias = tmp_keys[1]
                            try:
                                value = data_row[key]
                                if value is not None:
                                    keys[alias] = value
                                else:
                                    if debug:
                                        print(f"None value for key = {key}")
                                        print(data_row)
                                    keys = None
                            except:
                                if debug:
                                    print("-"*30)
                                    print(f"key = {key}")
                                    print(data_row)
                                keys = None
                                continue
                        # con malas llaves no se ppuede hacer el request
                        # y no se puede continuar en profundidad
                        if keys == None:
                            if debug:
                                print("/"*30)
                            continue
                        sub_tables_dict = table_relations["TABLES"]
                        if debug:
                            print(f"len subtables = {len(sub_tables_dict)}")
                        for sub_table_dict in sub_tables_dict:
                            sub_table = list(sub_table_dict.keys())[0]
                            if debug:
                                print(f"sub_table = {sub_table}")

                            try:
                                sub_table_data = []
                                tmp_data = self.utils.request_register(
                                    db, keys, sub_table)
                                for i, row in tmp_data.iterrows():
                                    req_data = self.request_graph(
                                        row, [sub_table_dict], None, debug)
                                    # sub_table_data.append({"table":sub_table,"data":req_data})
                                    sub_table_data.append(req_data)
                                ndata.append(
                                    {"table": sub_table, "data": sub_table_data, "keys": keys})

                            except:
                                if debug:
                                    print(data_row)
                                    print(
                                        f"db = {db} keys={keys} sub_table={sub_table}")
                                    print("|"*30)
                                continue
                else:
                    if debug:
                        print("*"*30)

            return ndata

    def graph2json(self, fields, regs, filter_function=None):
        """
        Recursive algorithm to parse the graph to a json structure.
        """
        output = {}
        if is_dict(regs):
            table = regs["table"]
            data = regs["data"]
            for i in data:
                value = {}
                if is_serie(i):
                    if table_exists(fields, table):
                        # this allows to jump relatioship tables that are not wanted
                        value = parse_table(fields, table, i, filter_function)
                        output.update(value)
                if is_list(i):
                    last_dict = {}
                    for j in i:
                        if is_serie(j):
                            value = parse_table(
                                fields, table, j, filter_function)

                        if is_dict(j):
                            last_dict = j
                            out = self.graph2json(fields, j, filter_function)
                            value.update(out)
                    if table_exists(fields, table):
                        section = fields[table]["alias"]
                        if section == "":
                            output.update(value)
                        else:
                            if section_exist(section, output.keys()):
                                output[section].append(value)
                            else:
                                output[section] = [value]
                    else:
                        if last_dict:
                            sub_table = last_dict["table"]
                            if table_exists(fields, sub_table):
                                section = fields[sub_table]["alias"]
                                if section == "":
                                    output.update(value)
                                else:
                                    if section_exist(section, output.keys()):
                                        if value:  # value !={}
                                            output[section].append(
                                                value[section][0])
                                    else:
                                        if value:
                                            output[section] = [
                                                value[section][0]]
                            else:
                                if section_exist("unkown", output.keys()):
                                    output["unkown"].append(value)
                                else:
                                    output["unkown"] = [value]
        else:
            for reg in regs:
                out = self.graph2json(fields, reg, filter_function)
                output.update(out)
        return output

    def parse_subsections(self, regs, graph_fields):
        """
        Method to parse subsections for multiple register gotten from the DB.
        """
        sub_section = {}
        for i in graph_fields.keys():
            alias = graph_fields[i]["alias"]
            if "sub_section" in graph_fields[i].keys():
                sub_section[alias] = graph_fields[i]["sub_section"]

        for i in range(len(regs)):
            old_reg = regs[i]
            new_reg = {}
            for j in old_reg.keys():
                if j in sub_section.keys():
                    if sub_section[j] in new_reg.keys():
                        new_reg[sub_section[j]].append({j: old_reg[j]})
                    else:
                        new_reg[sub_section[j]] = [{j: old_reg[j]}]
                else:
                    new_reg[j] = old_reg[j]
            regs[i] = new_reg
        return regs

    def parse_subsection(self, reg, sub_sections):
        """
        Allows to parce the subsection for an specific register from the DB.
        """
        new_reg = {}
        for j in reg.keys():
            if j in sub_sections.keys():
                if sub_sections[j] in new_reg.keys():
                    new_reg[sub_sections[j]].append({j: reg[j]})
                else:
                    new_reg[sub_sections[j]] = [{j: reg[j]}]
            else:
                new_reg[j] = reg[j]
        return new_reg

    def run_graph(self, data, graph_schema, max_threads=None, debug=False):
        """
        Allows to perform the relations for multiple registers in parallel.
        User this with careful, it's doesn´t support checkpoint and the regs are allocate in RAM memory,
        then this cna be expensive.
        """
        if max_threads is None:
            jobs = psutil.cpu_count()
        else:
            jobs = max_threads
        regs = []
        if jobs == 1:
            for i, row in tqdm(data.iterrows(), total=data.shape[0]):
                reg = self.request_graph(
                    row, graph_schema["GRAPH"], main_table=graph_schema["MAIN_TABLE"], debug=0)
                regs.append(reg)
        else:
            regs = Parallel(n_jobs=jobs, backend='threading', verbose=10)(delayed(self.request_graph)(
                row, graph_schema["GRAPH"], main_table=graph_schema["MAIN_TABLE"], debug=0) for i, row in data.iterrows())
        return regs

    def run_graph2json(self, regs, graph_fields, filter_function=None):
        """
        Allows to parse multiple registers from the graph to json and apply the filter function provided by the user.
        """
        output = []
        for reg in regs:
            out = self.graph2json(graph_fields, reg, filter_function)
            output.append(out)
        return output

    def request_graph2mongodb(self, dbclient, db_name, data_row, graph_schema, main_table, graph_fields, sub_sections, filter_function=None, checkpoint=None):
        """
        method to extract the data from Oracle and save the results in MongoDB.
        Allows to:
        * Walk the graph
        * Parse the graph to json
        * Paser subsection to put some information in a specific keys ex: for scienti the field details.
        * do checkpoints
        * save failed registers in a collection_failed.
        """
        tables = graph_schema["GRAPH"]
        try:
            reg = self.request_graph(data_row, tables, main_table)
            raw = self.graph2json(graph_fields, reg, filter_function)
            out = self.parse_subsection(raw, sub_sections)
            dbclient[db_name].get_collection(
                graph_fields[main_table]["alias"], codec_options=oracle_codec_options).insert_one(out)
            if checkpoint:
                ckp_info = graph_schema["CHECKPOINT"]
                reg = {}
                for key in ckp_info["KEYS"]:
                    reg[key] = data_row[key]
                checkpoint.update(
                    db_name, graph_fields[main_table]["alias"], reg)

        except:
            failed_collection = graph_fields[main_table]["alias"]+"_failed"
            print(
                f"Error parsing register, record added to the collection = {failed_collection} ")
            print(sys.exc_info())
            dbclient[db_name].get_collection(
                failed_collection, codec_options=oracle_codec_options).insert_one(data_row.to_dict())

    def run2mongodb(self, data, graph_schema, graph_fields, db_name, mongodb_uri="mongodb://localhost:27017/", max_threads=None, filter_function=None, checkpoint: UkuPachaCheckPoint = None):
        """
        Method to run all in parallel
        """
        sub_sections = {}
        for i in graph_fields.keys():
            alias = graph_fields[i]["alias"]
            if "sub_section" in graph_fields[i].keys():
                sub_sections[alias] = graph_fields[i]["sub_section"]
        dbclient = MongoClient(mongodb_uri)
        if max_threads is None:
            jobs = psutil.cpu_count()
        else:
            jobs = max_threads

        if jobs == 1:
            for i, row in tqdm(data.iterrows(), total=data.shape[0]):
                self.request_graph2mongodb(
                    dbclient, db_name, row, graph_schema, graph_schema["MAIN_TABLE"], graph_fields, sub_sections, filter_function)
        else:
            Parallel(n_jobs=jobs, backend='threading', verbose=10)(delayed(self.request_graph2mongodb)(
                dbclient, db_name, row, graph_schema, graph_schema["MAIN_TABLE"], graph_fields, sub_sections, filter_function, checkpoint) for i, row in data.iterrows())

    def save_json(self, output_file, data):
        """
        Method to save data to json file.
        """
        with open(output_file, 'w') as fp:
            json.dump(data, fp, cls=JsonEncoder, indent=4)

    def run2file(self, output_file, data, graph_schema, graph_fields, max_threads=None, debug=False, save_regs=False, save_raws=False, filter_function=None):
        regs = self.run_graph(data, graph_schema, max_threads, debug)
        if save_regs:
            self.save_json(output_file+".regs.json", regs)

        raws = self.run_graph2json(regs, graph_fields, filter_function)
        if save_raws:
            self.save_json(output_file+".raws.json", raws)

        output = self.parse_subsections(raws, graph_fields)
        self.save_json(output_file, output)
        print(f"Process finished, file {output_file} save")
